﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using Composite.Core.IO;
using Composite.Data;
using Composite.Data.Plugins.DataProvider;
using Microsoft.Practices.EnterpriseLibrary.Common.Configuration;
using Microsoft.Practices.EnterpriseLibrary.Common.Configuration.ObjectBuilder;
using Microsoft.Practices.ObjectBuilder;


namespace Composite.WritableDataProvider
{
    /// <summary>
    /// This is the implementation of the writable data provider
    /// Needed configuration = <add dataDirectory="~/App_Data/Composite/VCardDataStores" type="Composite.WritableDataProvider.VCardDataProvider, Composite.WritableDataProvider" name="VCardDataProvider" />
    /// 
    /// Here is how to add a new vCard
    /// <code>
    /// using (DataConnection connection = new DataConnection()) 
    /// {
    ///     IVCard card = connection.CreateNew<IVCard>();
    ///     card.Id = Guid.NewGuid();
    ///     card.Name = "MyName";
    ///     card.FormattedName = "My name, my Name";
    ///     card.Title = "My title";
    ///     card.EMail = "ljlj@com.com";
    ///     
    ///     connection.Add(card);
    /// }
    /// </code> 
    /// 
    /// Here is how to update a vCard
    /// <code>
    /// using (DataConnection connection = new DataConnection()) 
    /// {
    ///     IVCard updateCard = connection.Get<IVCard>().First();
    ///     updateCard.Name = "My New Name";
    ///     
    ///     connection.Update(updateCard);
    /// }
    /// </code> 
    /// 
    /// 
    /// Here is how to delete a vCard
    /// <code>
    /// using (DataConnection connection = new DataConnection()) 
    /// {
    ///     IVCard updateCard = connection.Get<IVCard>().First();
    ///     
    ///     connection.Delete(updateCard);
    /// }
    /// </code> 
    /// </summary>
    [ConfigurationElementType(typeof(VCardDataProviderData))]
    public class VCardDataProvider : IWritableDataProvider
    {
        private DataProviderContext _context;
        private readonly string _dataPath;


        /// <summary>
        /// 
        /// </summary>
        /// <param name="dataPath">The VCard files are stored here</param>
        public VCardDataProvider(string dataPath)
        {
            _dataPath = dataPath;
            if (!C1Directory.Exists(_dataPath))
            {
                C1Directory.CreateDirectory(_dataPath);
            }
        }



        /// <summary>
        /// This is provided by C1 and we use it for creating DataSourceId's
        /// </summary>
        public DataProviderContext Context
        {
            set { _context = value; }
        }



        /// <summary>
        /// In this example we only support one interface: IVCard
        /// </summary>
        /// <returns></returns>
        public IEnumerable<Type> GetSupportedInterfaces()
        {
            return new[] { typeof(IVCard) };
        }



        /// <summary>
        /// Find all vCards and return VCard objects as a queryable
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public IQueryable<T> GetData<T>() where T : class, IData
        {
            return
                C1Directory.GetFiles(_dataPath).
                Select(f => (T)CreateVCardFromPath(f)).
                AsQueryable();
        }



        /// <summary>
        /// This method finds a specific vCard by the stored id
        /// in the VCardDataId class.
        /// Important: The properties of the returned objects should
        /// be non-volatile! In other words. Let's say that I call this 
        /// two times and find the same object in both quieres like this:
        /// <code>
        /// using (DataConnection connection = new DataConnection())
        /// {
        ///     Guid myId; // Initialized
        ///     IVCard vCard1 = connection.Get<IVCard>().Where(f => f.Id == myId).Single();
        ///     IVCard vCard2 = connection.Get<IVCard>().Where(f => f.Id == myId).Single();
        /// 
        ///     vCard1.Name = "A new name";
        /// 
        ///     if (vCard1.Name == vCard2.Name) throw new Wrong();
        /// 
        ///     if (vCard1.Name != vCard2.Name) return OK!        
        /// }
        /// </code>
        /// 
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="dataId"></param>
        /// <returns></returns>
        public T GetData<T>(IDataId dataId) where T : class, IData
        {
            VCardDataId vCardDataId = (VCardDataId)dataId;

            string filePath =
                C1Directory.GetFiles(_dataPath).
                Where(f => GetIdFromFilePath(f) == vCardDataId.Id).
                Single();

            return (T)CreateVCardFromPath(filePath);
        }



        /// <summary>
        /// Save all new vCards.
        /// Note: This method does not handle if an id is already used or not.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="datas"></param>
        /// <returns></returns>
        public List<T> AddNew<T>(IEnumerable<T> datas) where T : class, IData
        {
            List<T> resultData = new List<T>();

            foreach (IVCard vCard in datas)
            {
                SaveVCard(vCard);

                // The type of vCard is not our own VCard. It is a generated class.
                // And we need to set the data source id, so we create a new instance.
                VCard newVCard = new VCard(_context.CreateDataSourceId(new VCardDataId { Id = vCard.Id }, typeof(IVCard)));
                newVCard.Name = vCard.Name;
                newVCard.FormattedName = vCard.FormattedName;
                newVCard.Title = vCard.Title;
                newVCard.EMail = vCard.EMail;
                resultData.Add((T)(IVCard)newVCard);
            }

            return resultData;
        }



        /// <summary>
        /// Save all updated vCards.
        /// </summary>
        /// <param name="datas"></param>
        public void Update(IEnumerable<IData> datas)
        {
            foreach (IVCard vCard in datas)
            {
                SaveVCard(vCard);
            }
        }



        /// <summary>
        /// Just delete the files. Then they will not be a part
        /// of the next query.
        /// </summary>
        /// <param name="dataSourceIds"></param>
        public void Delete(IEnumerable<DataSourceId> dataSourceIds)
        {
            foreach (DataSourceId dataSourceId in dataSourceIds)
            {
                VCardDataId vCardDataId = (VCardDataId)dataSourceId.DataId;

                string filePath = GetFilePath(vCardDataId.Id);

                if (C1File.Exists(filePath))
                {
                    C1File.Delete(filePath);
                }
            }
        }



        private string GetFilePath(IVCard vCard)
        {
            return GetFilePath(vCard.Id);
        }



        private string GetFilePath(Guid id)
        {
            return Path.Combine(_dataPath, string.Format("{0}.vcf", id));
        }



        private static Guid GetIdFromFilePath(string path)
        {
            return Guid.Parse(Path.GetFileNameWithoutExtension(path));
        }



        /// <summary>
        /// Deserialize a vCard from a file
        /// </summary>
        /// <param name="path"></param>
        /// <returns></returns>
        private IVCard CreateVCardFromPath(string path)
        {
            Guid id = GetIdFromFilePath(path);

            VCardDataId cardDataId = new VCardDataId
            {
                Id = id
            };

            VCard vCard = new VCard(_context.CreateDataSourceId(cardDataId, typeof(IVCard)));
            vCard.Id = id;
            vCard.Name = "";
            vCard.FormattedName = "";
            vCard.Title = "";
            vCard.EMail = "";

            string[] lines = File.ReadAllLines(path);
            foreach (string line in lines)
            {
                if (line.StartsWith("N:"))
                {
                    vCard.Name = line.Substring(2);
                }
                else if (line.StartsWith("FN:"))
                {
                    vCard.FormattedName = line.Substring(3);
                }
                else if (line.StartsWith("TITLE:"))
                {
                    vCard.Title = line.Substring(6);
                }
                else if (line.StartsWith("EMAIL;PREF;INTERNET:"))
                {
                    vCard.EMail = line.Substring(20);
                }
            }

            return vCard;
        }



        /// <summary>
        /// Serialize a vCard to a file
        /// </summary>
        /// <param name="vCard"></param>
        private void SaveVCard(IVCard vCard)
        {
            string filePath = GetFilePath(vCard);

            C1File.WriteAllLines(filePath, new[] { "BEGIN:VCARD", "VERSION:2.1" });
            C1File.AppendAllLines(filePath, new[] 
            { 
                "N:" + vCard.Name,
                "FN:" + vCard.FormattedName,
                "TITLE:" + vCard.Title,
                "EMAIL;PREF;INTERNET:" + vCard.EMail

            });

            C1File.AppendAllLines(filePath, new[] { "END:VCARD" });
        }
    }



    /// <summary>
    /// This class represents the configuration element in Composite.config
    /// We use this for configuring the folder that holds all the vCard files
    /// </summary>
    [Assembler(typeof(VCardDataProviderAssembler))]
    public sealed class VCardDataProviderData : DataProviderData
    {
        private const string _dataDirectoryPropertyName = "dataDirectory";
        [ConfigurationProperty(_dataDirectoryPropertyName, IsRequired = true)]
        public string DataDirectory
        {
            get { return (string)base[_dataDirectoryPropertyName]; }
            set { base[_dataDirectoryPropertyName] = value; }
        }
    }



    /// <summary>
    /// This class is used by the internals of C1 to create plugin objects with configuration
    /// We use this, given the configuration object, to create the data provider.
    /// </summary>
    public sealed class VCardDataProviderAssembler : IAssembler<IDataProvider, DataProviderData>
    {
        public IDataProvider Assemble(IBuilderContext context, DataProviderData objectConfiguration, IConfigurationSource configurationSource, ConfigurationReflectionCache reflectionCache)
        {
            VCardDataProviderData configuration = (VCardDataProviderData)objectConfiguration;

            string resolvedPath = PathUtil.Resolve(configuration.DataDirectory);
            return new VCardDataProvider(resolvedPath);
        }
    }
}
